static char rcsid[] = "@(#)$Id: header_alloc.c,v 1.10.8.6 2004/11/03 12:24:50 hurtta Exp $";


/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 1.10.8.6 $   $State: Exp $
 *
 *  Author: Kari Hurtta <hurtta+elm@posti.FMI.FI> (was hurtta+elm@ozone.FMI.FI)
 *****************************************************************************
 * Some code copied from src/newmbox.c. That file is following
 * copyright:
 *
 *  The Elm Mail System 
 *
 *			Copyright (c) 1988-1992 USENET Community Trust
 *			Copyright (c) 1986,1987 Dave Taylor
 *****************************************************************************/

#include "def_messages.h"

DEBUG_VAR(Debug,__FILE__,"messages");


void realloc_headers(storage,current_count)
     struct current_storage *storage;
     int current_count;
{

    /* Code (indirectly) copied from src/newmbox.c 
       function read_headers()
    */

    if (current_count >= storage->max_headers) {
	struct header_rec **new_headers;
	int new_max;

	new_max = storage->max_headers + KLICK;
	if (storage->max_headers == 0) {
	    new_headers = (struct header_rec **)
		safe_malloc(new_max * sizeof(struct header_rec *));
	} else {
	    new_headers = (struct header_rec **)
		safe_realloc((char *) storage->headers,
			     new_max * sizeof(struct header_rec *));
	}
	storage->headers = new_headers;
	while (storage->max_headers < new_max)
	    storage->headers[storage->max_headers++] = NULL;
    }
}

void alloc_header(storage,current_count)
     struct current_storage *storage;
     int current_count;
{
    if (storage->headers[current_count] == NULL) {
	struct header_rec *h = safe_malloc(sizeof(*h));
		
	header_zero(h);
	storage->headers[current_count] = h;
    }
}


static time_t now = 0;

void header_clear(h)      /* Free's fields */
     struct header_rec *h;
{
    if (0 == now) now = time(NULL);
    
    h->lines            = 0;
    h->index_number_X   = 0; 
    h->status           = 0;
    h->encrypted        = 0;
    h->exit_disposition = 0;
    h->status_chgd      = 0;
    h->content_length   = -1;
    h->offset           = -1;
    h->received_time    = now;
    h->env_from[0]      = '\0';
    if (h->from)
	free_addr_items(h->from);
    h->from             = NULL;
    if (h->to)
	free_addr_items(h->to);
    h->to               = NULL;
    if (h->cc)
	free_addr_items(h->cc);
    h->cc               = NULL;
    h->messageid[0]     = '\0';
    h->time_zone[0]     = '\0';
    h->time_menu[0]     = '\0';
    h->tz_offset        = 0;
    if (h->subject)
	free_string(&(h->subject));
    h->subject          = NULL;

    if (h->list_phrase)
	free_string(&(h->list_phrase));
    h->list_phrase      = NULL;
    if (h->list_id)
	free(h->list_id);
    h->list_id          = NULL;

    h->mailx_status[0]  = '\0';
    mime_t_clear (&(h->mime_rec));
    free_rec_mbx_info(h);
#ifdef USE_PGP
    h->pgp              = 0;
#endif
    h->binary           = 0;
    h->have_from        = 0;
    h->body_parsed      = 0;
    h->mime_parsed      = 0;
    h->header_charset   = NULL;
    h->override_charset = NULL;
    free_partial_vector(h);
}

void header_free(h)
     struct header_rec **h;
{
    header_clear(*h);
    free(*h);
    *h = NULL;
}

void header_zero(h)
     struct header_rec *h;
{
    if (0 == now) now = time(NULL);
    
    h->lines            = 0;
    h->index_number_X   = 0; 
    h->status           = 0;
    h->encrypted        = 0;
    h->exit_disposition = 0;
    h->status_chgd      = 0;
    h->content_length   = -1;
    h->offset           = -1;
    h->received_time    = now;
    h->time_sent        = 0;
    h->env_from[0]      = '\0';
    h->from             = NULL;
    h->to               = NULL;
    h->cc               = NULL;
    h->messageid[0]     = '\0';
    h->time_zone[0]     = '\0';
    h->time_menu[0]     = '\0';
    h->tz_offset        = 0;
    h->subject          = NULL;
    h->list_phrase      = NULL;
    h->list_id          = NULL;
    h->mailx_status[0]  = '\0';
    mime_t_zero (&(h->mime_rec));
    h->mbx_info         = NULL;
#ifdef USE_PGP
    h->pgp              = 0;
#endif
    h->binary           = 0;
    h->have_from        = 0;
    h->body_parsed      = 0;
    h->mime_parsed      = 0;
    h->header_charset   = NULL;
    h->override_charset = NULL;
    h->partial_vector   = NULL;
    h->partial_len      = 0;
}

static void parse_list_id P_((struct header_rec *current_header,
			      header_list_ptr tmphdr));
static void parse_list_id(current_header,tmphdr)
     struct header_rec *current_header;
     header_list_ptr tmphdr;
{
    char ** tokens = rfc822_tokenize(tmphdr->body);
    int i;
    char * phrase = NULL;
    char * id = NULL;

    /* 1) read phrase -- comments are not allowed */

    for (i = 0; tokens[i]; i++) {

	if ('\n' == tokens[i][0] &&
	    !tokens[i+1])
	    break;

	if ('<' == tokens[i][0])
	    break;

	if ('(' == tokens[i][0])
	    goto error;

	if ('\n' == tokens[i][0]) 
	    phrase = strmcat(phrase," ");	
	else
	    phrase = strmcat(phrase,tokens[i]);	
    }

    if (!tokens[i] || '<' != tokens[i][0])
	goto error;

    if (current_header->list_phrase || 
	current_header->list_id)
	goto error;

    if (phrase)
	current_header->list_phrase = 
	    hdr_to_string(HDR_PHRASE,phrase,
			  current_header->header_charset,
			  !(current_header -> status &
			    NOHDRENCODING) &&
			  is_rfc1522(phrase));

    /* 2) read id -- comments and whitespace is not allowed */

    for (i++; tokens[i]; i++) {

	if ('\n' == tokens[i][0] &&
	    !tokens[i+1])
	    break;

	if ('>' == tokens[i][0])
	    break;

	if ('(' == tokens[i][0])
	    goto error;
	if (' ' == tokens[i][0])
	    goto error;
	if ('\n' == tokens[i][0])
	    goto error;
	
	id = strmcat(id,tokens[i]);	
    }

    if (!tokens[i] || '>' != tokens[i][0])
	goto error;

    current_header->list_id = id; 
    id = NULL;

 error:
    if (phrase)
	free(phrase);
    if (id)
	free(id);

    free_rfc822tokenized(tokens);
}

void header_parse_helper(current_header,parsed_headers)
     struct header_rec *current_header;
     header_list_ptr parsed_headers;
{
    header_list_ptr tmphdr;

    if (NULL != (tmphdr = locate_header_by_name(parsed_headers,
						"Sensitivity"))) {
	if (0 == strcmp(tmphdr->body,"2"))
	    current_header->status |= PRIVATE_MAIL;
	else if (0 == strcmp(tmphdr->body,"3"))
	    current_header->status |= CONFIDENTIAL;
    }
    
    if (NULL != (tmphdr = locate_header_by_name(parsed_headers,
						"Importance"))) {
	if (0 == strcmp(tmphdr->body,"2"))
	    current_header->status |= URGENT;
    }
    
    if (NULL != (tmphdr = locate_header_by_name(parsed_headers,
						"Priority"))) {
	if (0 != istrcmp(tmphdr->body,"normal") &&
	    0 != istrcmp(tmphdr->body,"non-urgent"))
	    current_header->status |= URGENT;
    }

    if (NULL != (tmphdr = locate_header_by_name(parsed_headers,
						"Action"))) {
	current_header->status |= ACTION;
    }

    if (NULL != (tmphdr = locate_header_by_name(parsed_headers,
						"Content-Type"))) {
	if (0 == strcmp(tmphdr->body,"mailform"))
	    current_header->status |= FORM_LETTER | PRE_MIME_CONTENT;
	else if (is_pre_mime_content_type(&current_header->mime_rec,
					  tmphdr->body)) {
	    DPRINT(Debug,10,(&Debug, 
			     "NOT mime's content-type: %s\n",
			     tmphdr->body));
	    current_header->status |= PRE_MIME_CONTENT;
	}
    }
    
    if (current_header->binary) {
	DPRINT(Debug,10,(&Debug, 
			 "-- A binary message\n"));
    }
        
    if (req_mime_hdrencoding &&
	!(current_header->status & MIME_MESSAGE)) {
	current_header -> status |= NOHDRENCODING;
	DPRINT(Debug,10,(&Debug, 
			 "-- Turning on NOHDRENCODING\n"));
    }
    
    if (!req_mime_bodyencoding && 
	!(current_header->status & PRE_MIME_CONTENT)) {
	current_header -> status |= MIME_MESSAGE;
	DPRINT(Debug,10,(&Debug, 
			 "-- Turning on MIME_MESSAGE\n"));
    }

    if (NULL != (tmphdr = locate_header_by_name(parsed_headers,
						"Status"))) {
	strfcpy(current_header->mailx_status, tmphdr->body, WLEN);
	remove_possible_trailing_spaces(current_header->
					mailx_status);
	
	/* Okay readjust the status. If there's an 'R', message
	 * is read; if there is no 'R' but there is an 'O', message
	 * is unread. In any case it isn't new because a new 
	 * message wouldn't have a Status: header. 
	 * -- if message is re$ynced there may now be 
	 * Status: -header
	 */
	if (index(current_header->mailx_status, 'R') != NULL)
	    current_header->status &= ~(NEW | UNREAD);
	else if (index(current_header->mailx_status,'O') != NULL) {
	    current_header->status &= ~NEW;
	    current_header->status |= UNREAD;
	}
	if (index(current_header->mailx_status, 'r') != NULL)
	    current_header->status |= REPLIED_TO;
    }
    
    if (NULL != (tmphdr = locate_header_by_name(parsed_headers,
						"Subject"))) {
	char * Subject = safe_strdup(tmphdr->body);
	
	if (current_header->subject)
	    free_string(&(current_header->subject));
	
	remove_possible_trailing_spaces(Subject);
	
	current_header->subject = 
	    hdr_to_string(HDR_TEXT,
			  Subject,
			  current_header->header_charset,
			  !(current_header -> status &
			    NOHDRENCODING) &&
			  is_rfc1522(Subject));		
	
	/* This does not work correctly with unknown character 
	   sets... */
	remove_control(current_header->subject);
	
	free(Subject);
    } else if (!current_header->subject)
	current_header->subject = new_string(display_charset);
    

    if (NULL != (tmphdr = locate_header_by_name(parsed_headers,
						"List-Id"))) {
	parse_list_id(current_header,tmphdr);
    }
    
    for (tmphdr = locate_header_by_name(parsed_headers,
					"From");
	 tmphdr;
	 tmphdr = tmphdr -> next_this_header) {
	
	struct addr_item * addrs = 
	    break_down_address(tmphdr->body, 
			       !(current_header -> status & 
				 NOHDRENCODING) &&
			       is_rfc1522(tmphdr->body),
			       current_header->header_charset);
	
	if (addrs) {	    

	    if (current_header->from) {
		append_addr_items(& (current_header->from), addrs);
		free_addr_items(addrs);
	    
	    } else
		current_header->from = addrs;

	}	       
    }

    for (tmphdr = locate_header_by_name(parsed_headers,
					"To");
	 tmphdr;
	 tmphdr = tmphdr -> next_this_header) {

	struct addr_item * addrs = 
	    break_down_address(tmphdr->body,
			       !(current_header -> status & 
				 NOHDRENCODING) &&
			       is_rfc1522(tmphdr->body),
			       current_header->header_charset);
	
	if (addrs) {

	    if (current_header->to) {
		append_addr_items(& (current_header->to), addrs);
		free_addr_items(addrs);
	    } else
		current_header->to = addrs;
	}	       
    }

    for (tmphdr = locate_header_by_name(parsed_headers,
					"Cc");
	 tmphdr;
	 tmphdr = tmphdr -> next_this_header) {

	struct addr_item * addrs = 
	    break_down_address(tmphdr->body,
				       !(current_header -> status & 
					 NOHDRENCODING) &&
			       is_rfc1522(tmphdr->body),
			       current_header->header_charset);
	
	if (addrs) {
	    if (current_header->cc) {
		append_addr_items(& (current_header->cc), addrs);
		free_addr_items(addrs);
	    } else 
		current_header->cc = addrs;
	}
    }

    /* message_hide_hack values:   0 == none                          */
                                /* 1 = FOLDER INTERNAL DATA           */

    switch (message_hide_hack) {
	header_list_ptr tmphdr2;

    case 1:
	tmphdr  = locate_header_by_name(parsed_headers, "Subject");
	tmphdr2 = locate_header_by_name(parsed_headers, "X-IMAP");

	if (tmphdr && tmphdr2 &&
	    NULL != strstr(tmphdr->body,
			   "DON'T DELETE THIS MESSAGE -- FOLDER INTERNAL DATA")) {
	    current_header->exit_disposition = HIDE;
	    DPRINT(Debug,10,(&Debug, "message_hide_hack=%d -- hiding message: %s\n",
			     message_hide_hack,tmphdr->body));
	}
	break;
    }
}			      

void start_body_helper(current_header,content_start,parsed_headers)
     struct header_rec *current_header;
     long content_start;
     header_list_ptr parsed_headers;
{
    /* Notice that mark_keep_folder() also uses mime_rec.offset */
    current_header->mime_rec.offset = content_start;
    DPRINT(Debug,10,(&Debug, 
		     "-- content_start=%ld\n",content_start));

    if (0 != (current_header -> status & MIME_MESSAGE)) {
	long part_offset = current_header->mime_rec.begin_offset; /* Sic! */
	long body_offset = content_start;
	
	parse_mime_headers1(&current_header->mime_rec,
			    parsed_headers,
			    part_offset,body_offset,0,
			    current_header->header_charset);

	if (0 != (get_type_flags(current_header->mime_rec.TYPE) & 
		  MIME_SIGNED)) {
#ifdef USE_PGP
	    char protocol[STRING];
	    if (mime_get_param ("protocol", protocol, 
				current_header->
				mime_rec.type_opts, 
				sizeof protocol) &&
		0 == istrcmp(protocol,
			     "application/pgp-signature"))
		current_header->pgp |= PGP_SIGNED_MESSAGE;
#endif
	}
    } else {
	/* Notice that mark_keep_folder() also uses mime_rec.offset */
	current_header->mime_rec.offset = content_start;
	DPRINT(Debug,10,(&Debug, 
			 "-- content_start=%ld\n",content_start));
    }
    
}

/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 *  buffer-file-coding-system: iso-8859-1
 * End:
 */
